import os
import json
from http import HTTPStatus
from unittest.mock import Mock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from sanic.request import Request
from sanic import response
from tests.unit import conftest
from tests.unit.conftest import insert_test_user, TEST_USER_ADMIN
from alembic.runtime.migration import MigrationContext

import rasax.community.database.utils as db_utils
from rasax.community.database.admin import Role
import rasax.community.sql_migrations as migrations

import sqlalchemy.dialects
import sqlalchemy.exc
from sqlalchemy.orm import Session


# Overwrite fixture to run per function
@pytest.fixture()
def session():
    session_maker = conftest.session_maker_for_tests()
    session = session_maker()

    migrations.run_migrations(session)

    # Insert general test user which can be used in later tests
    insert_test_user(session, TEST_USER_ADMIN)

    return session_maker()


def test_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=True)
    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute(f"DROP table {Role.__tablename__};")


def test_is_db_revision_latest(session: Session):
    context = MigrationContext.configure(session.connection())
    heads = context.get_current_heads()
    current = context.get_current_revision()

    assert heads[0] == current
    assert db_utils.is_db_revision_latest(session)


def test_is_db_revision_latest_false(session_without_migration: Session):
    context = MigrationContext.configure(session_without_migration.connection())
    heads = context.get_current_heads()
    current = context.get_current_revision()

    assert not heads or heads[0] != current
    assert not db_utils.is_db_revision_latest(session_without_migration)


def test_turn_off_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=False)
    if isinstance(db_connection.dialect, sqlalchemy.dialects.sqlite.dialect):
        db_connection.execute(f"DROP table {Role.__tablename__};")


def test_turn_off_then_on_foreign_key_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=False)
    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=True)

    with pytest.raises(sqlalchemy.exc.IntegrityError):
        db_connection.execute(f"DROP table {Role.__tablename__};")


# noinspection PyProtectedMember
def test_db_query_from_environment():
    db_query_env_var = "DB_QUERY"

    # env var not set
    os.environ.pop(db_query_env_var, None)
    assert not db_utils._sql_query_parameters_from_environment()

    # env var correctly set
    query = {"driver": "some driver"}
    os.environ[db_query_env_var] = json.dumps(query)
    assert db_utils._sql_query_parameters_from_environment() == query

    # malformed env var
    malformed_db_query = '{"query": "badly formatted}'
    os.environ[db_query_env_var] = malformed_db_query
    assert not db_utils._sql_query_parameters_from_environment()


async def test_remove_session_bad_request_no_db_commit_called(monkeypatch: MonkeyPatch):

    db_session_mock = Mock()

    http_request = Request(None, None, None, None, None, None)
    http_response = response.json([], status=HTTPStatus.BAD_REQUEST)
    monkeypatch.setattr(Request, "get", lambda x, y: db_session_mock)

    await db_utils.remove_session(http_request, http_response)
    assert not db_session_mock.commit.called
    assert db_session_mock.close.called


async def test_remove_session_request_ok_db_commit_called(monkeypatch: MonkeyPatch):

    db_session_mock = Mock()

    http_request = Request(None, None, None, None, None, None)
    http_response = response.json([], status=HTTPStatus.OK)
    monkeypatch.setattr(Request, "get", lambda x, y: db_session_mock)

    await db_utils.remove_session(http_request, http_response)
    assert db_session_mock.commit.called
    assert db_session_mock.close.called
