from typing import Text, Any, Type

import pytest
from _pytest.monkeypatch import MonkeyPatch
from sqlalchemy.orm import Session

from rasax.community.database.admin import ConfigValue
from rasax.community.services.config_service import (
    ConfigService,
    MissingConfigValue,
    InvalidConfigValue,
    _CONFIG_SERVICE_DEFAULTS,
)
from tests.unit.conftest import reset_config_service


@pytest.fixture(autouse=True)
def clean_config_service(config_service: ConfigService) -> None:
    # Clean before testing
    reset_config_service(config_service)
    yield
    # Clean up after testing (in case other tests use the service without resetting it)
    reset_config_service(config_service)


def test_default_values(config_service: ConfigService):
    for key, value in _CONFIG_SERVICE_DEFAULTS.items():
        assert config_service.get_value(key) == value


def test_get_value_env_override(
    config_service: ConfigService, monkeypatch: MonkeyPatch
):
    config_service.set_value("KEY", "initial value")
    assert config_service.get_value("KEY") == "initial value"

    monkeypatch.setenv("KEY", "other value")
    assert config_service.get_value("KEY") == "other value"


def test_missing_key(config_service: ConfigService):
    with pytest.raises(MissingConfigValue):
        config_service.get_value("foobar")


def test_json_deserialization_error(session: Session, config_service: ConfigService):
    session.add(ConfigValue(key="foo", value="bar"))
    with pytest.raises(InvalidConfigValue):
        config_service.get_value("foo")


@pytest.mark.parametrize("value", [1, "test-string", True, False, None])
def test_set_value(value: Any, config_service: ConfigService):
    config_service.set_value("key", value)
    assert config_service.get_value("key") == value


@pytest.mark.parametrize("value", [..., set(), object()])
def test_set_value_error(value: Any, config_service: ConfigService):
    with pytest.raises(InvalidConfigValue):
        config_service.set_value("foobar", value)


@pytest.mark.parametrize(
    "value, expected_type", [(True, bool), (1, int), ("hello", str)],
)
def test_get_value_type(value: Any, expected_type: Type, config_service: ConfigService):
    config_service.set_value("key", value)
    config_service.get_value("key", expected_type=expected_type)


@pytest.mark.parametrize(
    "value, expected_type",
    [("true", bool), ("1", int), ("hello", str), ('"test"', str)],
)
def test_get_value_type_from_env(
    value: Text,
    expected_type: Type,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # When reading from environment variables, values are intepreted as JSON or
    # strings if deserialization fails
    monkeypatch.setenv("key", value)
    config_service.get_value("key", expected_type=expected_type)


def test_get_value_type_error(config_service: ConfigService):
    with pytest.raises(InvalidConfigValue):
        config_service.set_value("key", 123)
        config_service.get_value("key", expected_type=str)


@pytest.mark.parametrize(
    "value, expected",
    [("true", True), ("foo", "foo"), ("1", 1), ("null", None), ('"hello"', "hello")],
)
def test_get_value_env(
    value: Text, expected: Any, config_service: ConfigService, monkeypatch: MonkeyPatch
):
    # ConfigService should try to JSON-deserialize env variables if possible,
    # or return them as a string if that was not possible
    monkeypatch.setenv("key", value)
    assert config_service.get_value("key") == expected
